@@ -3,35 +3,38 @@
  *  Module    : tinrc.h
  *  Author    : Jason Faultless <jason@altarstone.com>
  *  Created   : 1999-04-13
- *  Updated   : 2017-08-13
+ *  Updated   : 2022-09-19
  *  Notes     :
  *
- * Copyright (c) 1999-2019 Jason Faultless <jason@altarstone.com>
+ * Copyright (c) 1999-2023 Jason Faultless <jason@altarstone.com>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
- * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer.
+ *
+ * 1. Redistributions of source code must retain the above copyright notice,
+ *    this list of conditions and the following disclaimer.
+ *
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
- * 3. The name of the author may not be used to endorse or promote
- *    products derived from this software without specific prior written
- *    permission.
  *
- * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
- * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
- * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
- * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
- * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
- * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
- * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
- * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
- * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ * 3. Neither the name of the copyright holder nor the names of its
+ *    contributors may be used to endorse or promote products derived from
+ *    this software without specific prior written permission.
+ *
+ * * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+ * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
+ * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
+ * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
+ * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
+ * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
+ * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
+ * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
+ * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+ * POSSIBILITY OF SUCH DAMAGE.
  */
 
 
@@ -53,15 +56,27 @@ struct t_config {
 	/*
 	 * Chars used to show article status
 	 */
-	char art_marked_deleted;
-	char art_marked_inrange;	/* misnomer, as it's also used for other ranges */
-	char art_marked_return;
-	char art_marked_selected;
-	char art_marked_recent;
-	char art_marked_unread;
-	char art_marked_read;
-	char art_marked_killed;
-	char art_marked_read_selected;
+#	if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+		wchar_t art_marked_deleted;
+		wchar_t art_marked_inrange;	/* misnomer, as it's also used for other ranges */
+		wchar_t art_marked_return;
+		wchar_t art_marked_selected;
+		wchar_t art_marked_recent;
+		wchar_t art_marked_unread;
+		wchar_t art_marked_read;
+		wchar_t art_marked_killed;
+		wchar_t art_marked_read_selected;
+#	else
+		char art_marked_deleted;
+		char art_marked_inrange;	/* misnomer, as it's also used for other ranges */
+		char art_marked_return;
+		char art_marked_selected;
+		char art_marked_recent;
+		char art_marked_unread;
+		char art_marked_read;
+		char art_marked_killed;
+		char art_marked_read_selected;
+#	endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 	char editor_format[PATH_LEN];		/* editor + parameters  %E +%N %F */
 	char default_goto_group[HEADER_LEN];		/* default for the 'g' command */
 	char default_mail_address[HEADER_LEN];
@@ -122,6 +137,9 @@ struct t_config {
 	char verbatim_end_regex[LEN];			/* regex used to find the end of a verbatim block */
 	char savedir[PATH_LEN];				/* directory to save articles to */
 	char spamtrap_warning_addresses[LEN];
+#	ifdef NNTPS_ABLE
+		char tls_ca_cert_file[PATH_LEN];	/* file containing trusted CA certificates */
+#	endif /* NNTPS_ABLE */
 	char url_handler[LEN];				/* Helper app for opening URL's */
 	char xpost_quote_format[LEN];
 	int filter_days;					/* num of days an article filter can be active */
@@ -155,6 +173,7 @@ struct t_config {
 	int score_kill;						/* default score for "kill" filter rules */
 	int score_select;					/* default score for "hot" filter rules */
 	int trim_article_body;				/* remove unnecessary blank lines */
+	int show_help_mail_sign;			/* show help/mail sign in level title */
 #	ifdef HAVE_COLOR
 		int col_back;						/* standard background color */
 		int col_from;						/* color of sender (From:) */
@@ -174,7 +193,7 @@ struct t_config {
 		int col_quote2;						/* color of twice quoted lines */
 		int col_quote3;						/* color of >=3 times quoted lines */
 		int col_extquote;					/* color of quoted external text */
-		int col_response;					/* color of respone counter */
+		int col_response;					/* color of response counter */
 		int col_signature;					/* color of signature */
 		int col_urls;						/* color of urls highlight */
 		int col_verbatim;					/* color of verbatim blocks */
@@ -200,7 +219,6 @@ struct t_config {
 	int auto_cc_bcc;					/* add your name to cc/bcc automatically */
 	t_bool auto_list_thread;			/* list thread when entering it using right arrow */
 	t_bool auto_reconnect;				/* automatically reconnect to news server */
-	t_bool auto_save;					/* save thread with name from Archive-name: field */
 	t_bool batch_save;					/* save arts if -M/-S command line switch specified */
 	t_bool beginner_level;				/* beginner level (shows mini help a la elm) */
 	t_bool cache_overview_files;		/* create local index files for NNTP overview files */
@@ -222,7 +240,7 @@ struct t_config {
 #	ifndef DISABLE_PRINTING
 		t_bool print_header;				/* print all of mail header or just Subject: & From lines */
 #	endif /* !DISABLE_PRINTING */
-	t_bool process_only_unread;			/* save/print//mail/pipe unread/all articles */
+	t_bool process_only_unread;			/* save/print/mail/pipe unread/all articles */
 	t_bool prompt_followupto;			/* display empty Followup-To header in editor */
 	int quote_style;					/* quoting behaviour */
 	t_bool show_description;
@@ -231,25 +249,27 @@ struct t_config {
 	t_bool show_signatures;				/* show signatures when displaying articles */
 	t_bool sigdashes;					/* set TRUE to prepend every signature with dashes */
 	t_bool signature_repost;			/* set TRUE to add signature when reposting articles */
-	t_bool start_editor_offset;
-#ifndef USE_CURSES
+#	ifndef USE_CURSES
 		t_bool strip_blanks;
-#endif /* !USE_CURSES */
+#	endif /* !USE_CURSES */
 	t_bool strip_newsrc;
+#	if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+		t_bool suppress_soft_hyphens;	/* set TRUE to remove soft hyphens (U+00AD) from articles */
+#	endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 	t_bool tex2iso_conv;			/* convert "a to Umlaut-a */
 	t_bool thread_catchup_on_exit;		/* catchup thread with left arrow key or not */
 	t_bool unlink_article;
-#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+#	if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
 		t_bool utf8_graphics;				/* use utf-8 characters for line drawing */
-#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
+#	endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 	t_bool verbatim_handling;			/* Detection of verbatim blocks */
 #	ifdef HAVE_COLOR
 		t_bool extquote_handling;		/* Detection of quoted text from external sources */
 #	endif /* HAVE_COLOR */
 	char inews_prog[PATH_LEN];
-#ifdef USE_CANLOCK
+#	ifdef USE_CANLOCK
 		int cancel_lock_algo;			/* algorithm used for Cancel-Lock/Cancel-Key */
-#endif /* USE_CANLOCK */
+#	endif /* USE_CANLOCK */
 	int interactive_mailer;			/* invoke user's mailreader */
 	t_bool use_mouse;					/* enables/disables mouse support under xterm */
 #	ifdef HAVE_KEYPAD
@@ -310,13 +330,13 @@ struct t_config {
 	int attrib_quick_kill_header;
 	int attrib_quick_select_header;
 	int attrib_mail_mime_encoding;
-#if defined(HAVE_ALARM) && defined(SIGALRM)
+#	if defined(HAVE_ALARM) && defined(SIGALRM)
 		/*
 		 * # seconds after which a read from the NNTP will timeout
 		 * NB: This is different from the NNTP server timing us out due to inactivity
 		 */
 		int nntp_read_timeout_secs;
-#endif /* HAVE_ALARM && SIGALRM */
+#	endif /* HAVE_ALARM && SIGALRM */
 	int attrib_post_mime_encoding;
 	int attrib_post_process_type;
 	int attrib_show_author;
@@ -329,7 +349,6 @@ struct t_config {
 	t_bool attrib_alternative_handling;
 	t_bool attrib_auto_list_thread;
 	t_bool attrib_auto_select;
-	t_bool attrib_auto_save;
 	t_bool attrib_batch_save;
 	t_bool attrib_delete_tmp_files;
 	t_bool attrib_group_catchup_on_exit;
@@ -349,7 +368,9 @@ struct t_config {
 	t_bool attrib_show_signatures;
 	t_bool attrib_sigdashes;
 	t_bool attrib_signature_repost;
-	t_bool attrib_start_editor_offset;
+#	if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+		t_bool attrib_suppress_soft_hyphens;
+#	endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 	t_bool attrib_tex2iso_conv;
 	t_bool attrib_thread_catchup_on_exit;
 	t_bool attrib_verbatim_handling;
