@@ -3,35 +3,38 @@
  *  Module    : config.c
  *  Author    : I. Lea
  *  Created   : 1991-04-01
- *  Updated   : 2016-04-17
+ *  Updated   : 2022-09-19
  *  Notes     : Configuration file routines
  *
- * Copyright (c) 1991-2016 Iain Lea <iain@bricbrac.de>
+ * Copyright (c) 1991-2023 Iain Lea <iain@bricbrac.de>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
- * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer.
+ *
+ * 1. Redistributions of source code must retain the above copyright notice,
+ *    this list of conditions and the following disclaimer.
+ *
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
- * 3. The name of the author may not be used to endorse or promote
- *    products derived from this software without specific prior written
- *    permission.
  *
- * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
- * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
- * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
- * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
- * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
- * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
- * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
- * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
- * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ * 3. Neither the name of the copyright holder nor the names of its
+ *    contributors may be used to endorse or promote products derived from
+ *    this software without specific prior written permission.
+ *
+ * * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+ * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
+ * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
+ * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
+ * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
+ * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
+ * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
+ * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
+ * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+ * POSSIBILITY OF SUCH DAMAGE.
  */
 
 
@@ -53,15 +56,33 @@
  */
 static t_bool match_item(char *line, const char *pat, char *dst, size_t dstlen);
 static t_bool rc_update(FILE *fp);
-static t_bool rc_post_update(FILE *fp);
+static t_bool rc_post_update(FILE *fp/* , struct t_version *upgrade */);
 static void write_server_config(void);
 #ifdef HAVE_COLOR
 	static t_bool match_color(char *line, const char *pat, int *dst, int max);
 #endif /* HAVE_COLOR */
 
 
-#define DASH_TO_SPACE(mark)	((char) (mark == '_' ? ' ' : mark))
-#define SPACE_TO_DASH(mark)	((char) (mark == ' ' ? '_' : mark))
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+#	define DASH_TO_SPACE(mark)	((wchar_t) (mark == L'_' ? L' ' : mark))
+#	define SPACE_TO_DASH(mark)	((wchar_t) (mark == L' ' ? L'_' : mark))
+#	define SET_RC_VAL(buf, rcval, defval) do { \
+		wchar_t *wbuf; \
+		if ((wbuf = char2wchar_t(buf))) { \
+			wbuf[1] = (wchar_t) '\0'; \
+			rcval = !wbuf[0] ? (wchar_t) defval : DASH_TO_SPACE(wbuf[0]); \
+			if (art_mark_width < wcswidth(wbuf, 1)) \
+				art_mark_width = wcswidth(wbuf, 1); \
+			free(wbuf); \
+		} \
+	} while (0)
+#else
+#	define DASH_TO_SPACE(mark)	((char) (mark == '_' ? ' ' : mark))
+#	define SPACE_TO_DASH(mark)	((char) (mark == ' ' ? '_' : mark))
+#	define SET_RC_VAL(buf, rcval, defval) do { \
+		rcval = !buf[0] ? defval : DASH_TO_SPACE(buf[0]); \
+	} while (0)
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 
 
 /*
@@ -74,7 +95,7 @@ read_config_file(
 {
 	FILE *fp;
 	char buf[LEN], tmp[LEN];
-	enum rc_state upgrade = RC_CHECK;
+	struct t_version *upgrade = NULL;
 #ifdef CHARSET_CONVERSION
 	int i;
 	t_bool is_7bit;
@@ -90,17 +111,17 @@ read_config_file(
 		if (buf[0] == '\n')
 			continue;
 		if (buf[0] == '#') {
-			if (upgrade == RC_CHECK && !global_file && match_string(buf, "# tin configuration file V", NULL, 0)) {
+			if (upgrade == NULL && !global_file && match_string(buf, "# tin configuration file V", NULL, 0)) {
 				upgrade = check_upgrade(buf, "# tin configuration file V", TINRC_VERSION);
-				if (upgrade != RC_IGNORE)
-					upgrade_prompt_quit(upgrade, CONFIG_FILE);
-				if (upgrade == RC_UPGRADE)
+				if (upgrade->state != RC_IGNORE)
+					upgrade_prompt_quit(upgrade, file); /* CONFIG_FILE */
+				if (upgrade->state == RC_UPGRADE)
 					rc_update(fp);
 			}
 			continue;
 		}
 
-		switch (tolower((unsigned char) buf[0])) {
+		switch (my_tolower((unsigned char) buf[0])) {
 		case 'a':
 			if (match_boolean(buf, "abbreviate_groupname=", &tinrc.abbreviate_groupname))
 				break;
@@ -115,47 +136,47 @@ read_config_file(
 				break;
 
 			if (match_string(buf, "art_marked_deleted=", tmp, sizeof(tmp))) {
-				tinrc.art_marked_deleted = !tmp[0] ? ART_MARK_DELETED : DASH_TO_SPACE(tmp[0]);
+				SET_RC_VAL(tmp, tinrc.art_marked_deleted, ART_MARK_DELETED);
 				break;
 			}
 
 			if (match_string(buf, "art_marked_inrange=", tmp, sizeof(tmp))) {
-				tinrc.art_marked_inrange = !tmp[0] ? MARK_INRANGE : DASH_TO_SPACE(tmp[0]);
+				SET_RC_VAL(tmp, tinrc.art_marked_inrange, MARK_INRANGE);
 				break;
 			}
 
 			if (match_string(buf, "art_marked_killed=", tmp, sizeof(tmp))) {
-				tinrc.art_marked_killed = !tmp[0] ? ART_MARK_KILLED : DASH_TO_SPACE(tmp[0]);
+				SET_RC_VAL(tmp, tinrc.art_marked_killed, ART_MARK_KILLED);
 				break;
 			}
 
 			if (match_string(buf, "art_marked_read=", tmp, sizeof(tmp))) {
-				tinrc.art_marked_read = !tmp[0] ? ART_MARK_READ : DASH_TO_SPACE(tmp[0]);
+				SET_RC_VAL(tmp, tinrc.art_marked_read, ART_MARK_READ);
 				break;
 			}
 
 			if (match_string(buf, "art_marked_read_selected=", tmp, sizeof(tmp))) {
-				tinrc.art_marked_read_selected = !tmp[0] ? ART_MARK_READ_SELECTED : DASH_TO_SPACE(tmp[0]);
+				SET_RC_VAL(tmp, tinrc.art_marked_read_selected, ART_MARK_READ_SELECTED);
 				break;
 			}
 
 			if (match_string(buf, "art_marked_recent=", tmp, sizeof(tmp))) {
-				tinrc.art_marked_recent = !tmp[0] ? ART_MARK_RECENT : DASH_TO_SPACE(tmp[0]);
+				SET_RC_VAL(tmp, tinrc.art_marked_recent, ART_MARK_RECENT);
 				break;
 			}
 
 			if (match_string(buf, "art_marked_return=", tmp, sizeof(tmp))) {
-				tinrc.art_marked_return = !tmp[0] ? ART_MARK_RETURN : DASH_TO_SPACE(tmp[0]);
+				SET_RC_VAL(tmp, tinrc.art_marked_return, ART_MARK_RETURN);
 				break;
 			}
 
 			if (match_string(buf, "art_marked_selected=", tmp, sizeof(tmp))) {
-				tinrc.art_marked_selected = !tmp[0] ? ART_MARK_SELECTED : DASH_TO_SPACE(tmp[0]);
+				SET_RC_VAL(tmp, tinrc.art_marked_selected, ART_MARK_SELECTED);
 				break;
 			}
 
 			if (match_string(buf, "art_marked_unread=", tmp, sizeof(tmp))) {
-				tinrc.art_marked_unread = !tmp[0] ? ART_MARK_UNREAD : DASH_TO_SPACE(tmp[0]);
+				SET_RC_VAL(tmp, tinrc.art_marked_unread, ART_MARK_UNREAD);
 				break;
 			}
 
@@ -171,8 +192,12 @@ read_config_file(
 			if (match_boolean(buf, "auto_reconnect=", &tinrc.auto_reconnect))
 				break;
 
-			if (match_boolean(buf, "auto_save=", &tinrc.auto_save))
-				break;
+			if (upgrade && upgrade->file_version < 10318) {
+				t_bool ignore;
+				/* option removed */
+				if (match_boolean(buf, "auto_save=", &ignore))
+					break;
+			}
 
 			break;
 
@@ -189,6 +214,11 @@ read_config_file(
 			if (match_boolean(buf, "cache_overview_files=", &tinrc.cache_overview_files))
 				break;
 
+#ifdef USE_CANLOCK
+			if (match_list(buf, "cancel_lock_algo=", txt_cancel_lock_algos, &tinrc.cancel_lock_algo))
+				break;
+#endif /* USE_CANLOCK */
+
 			if (match_boolean(buf, "catchup_read_groups=", &tinrc.catchup_read_groups))
 				break;
 
@@ -223,10 +253,8 @@ read_config_file(
 			if (match_color(buf, "col_quote3=", &tinrc.col_quote3, MAX_COLOR))
 				break;
 
-#ifdef HAVE_COLOR
 			if (match_color(buf, "col_extquote=", &tinrc.col_extquote, MAX_COLOR))
 				break;
-#endif /* HAVE_COLOR */
 
 			if (match_color(buf, "col_head=", &tinrc.col_head, MAX_COLOR))
 				break;
@@ -270,8 +298,14 @@ read_config_file(
 			if (match_color(buf, "col_markstroke=", &tinrc.col_markstroke, MAX_COLOR))
 				break;
 #endif /* HAVE_COLOR */
-			if (match_list(buf, "confirm_choice=", txt_confirm_choices, &tinrc.confirm_choice))
-				break;
+
+			if (upgrade && upgrade->file_version < 10316) {
+				if (match_list(buf, "confirm_choice=", txt_confirm_choices, &tinrc.confirm_choice))
+					break;
+			} else {
+				if (match_integer(buf, "confirm_choice=", &tinrc.confirm_choice, TINRC_CONFIRM_MAX))
+					break;
+			}
 
 			break;
 
@@ -429,7 +463,7 @@ read_config_file(
 			if (match_string(buf, "inews_prog=", tinrc.inews_prog, sizeof(tinrc.inews_prog)))
 				break;
 
-			if (match_integer(buf, "interactive_mailer=", &tinrc.interactive_mailer, INTERACTIVE_NONE))
+			if (match_integer(buf, "interactive_mailer=", &tinrc.interactive_mailer, (int) INTERACTIVE_NONE))
 				break;
 
 			break;
@@ -526,13 +560,8 @@ read_config_file(
 #endif /* HAVE_ALARM && SIGALRM */
 
 #ifdef HAVE_UNICODE_NORMALIZATION
-#	if (HAVE_UNICODE_NORMALIZATION >= 2)
-			if (match_integer(buf, "normalization_form=", &tinrc.normalization_form, NORMALIZE_NFD))
+			if (match_integer(buf, "normalization_form=", &tinrc.normalization_form, (int) NORMALIZE_MAX))
 				break;
-#	else
-			if (match_integer(buf, "normalization_form=", &tinrc.normalization_form, NORMALIZE_NFKC))
-				break;
-#	endif /* HAVE_UNICODE_NORMALIZATION >= 2 */
 #endif /* HAVE_UNICODE_NORMALIZATION */
 
 			break;
@@ -574,6 +603,16 @@ read_config_file(
 
 		case 'q':
 			if (match_string(buf, "quote_chars=", tinrc.quote_chars, sizeof(tinrc.quote_chars))) {
+				if (upgrade && upgrade->file_version < 10317) { /* %s/%S changed to %I */
+					char *q = tinrc.quote_chars;
+
+					while (*q) {
+						if (*q == '%' && (*(q + 1) == 's' || *(q + 1) == 'S'))
+							*(++q) = 'I';
+
+						q++;
+					}
+				}
 				quote_dash_to_space(tinrc.quote_chars);
 				break;
 			}
@@ -647,6 +686,9 @@ read_config_file(
 				break;
 			}
 
+			if (match_integer(buf, "show_help_mail_sign=", &tinrc.show_help_mail_sign, SHOW_SIGN_BOTH))
+				break;
+
 			if (match_boolean(buf, "show_only_unread_arts=", &tinrc.show_only_unread_arts))
 				break;
 
@@ -665,9 +707,6 @@ read_config_file(
 			if (match_string(buf, "spamtrap_warning_addresses=", tinrc.spamtrap_warning_addresses, sizeof(tinrc.spamtrap_warning_addresses)))
 				break;
 
-			if (match_boolean(buf, "start_editor_offset=", &tinrc.start_editor_offset))
-				break;
-
 			if (match_integer(buf, "sort_article_type=", &tinrc.sort_article_type, SORT_ARTICLES_BY_LINES_ASCEND))
 				break;
 
@@ -712,6 +751,11 @@ read_config_file(
 			if (match_string(buf, "strip_was_regex=", tinrc.strip_was_regex, sizeof(tinrc.strip_was_regex)))
 				break;
 
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+			if (match_boolean(buf, "suppress_soft_hyphens=", &tinrc.suppress_soft_hyphens))
+				break;
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
+
 			break;
 
 		case 't':
@@ -733,6 +777,11 @@ read_config_file(
 			if (match_boolean(buf, "thread_catchup_on_exit=", &tinrc.thread_catchup_on_exit))
 				break;
 
+#ifdef NNTPS_ABLE
+			if (match_string(buf, "tls_ca_cert_file=", tinrc.tls_ca_cert_file, sizeof(tinrc.tls_ca_cert_file)))
+				break;
+#endif /* NNTPS_ABLE */
+
 #if defined(HAVE_ICONV_OPEN_TRANSLIT) && defined(CHARSET_CONVERSION)
 			if (match_boolean(buf, "translit=", &tinrc.translit))
 				break;
@@ -766,7 +815,7 @@ read_config_file(
 
 #ifdef HAVE_COLOR
 			if (match_boolean(buf, "use_color=", &tinrc.use_color)) {
-				use_color = cmdline.args & CMDLINE_USE_COLOR ? bool_not(tinrc.use_color) : tinrc.use_color;
+				use_color = (cmdline.args & CMDLINE_USE_COLOR) ? bool_not(tinrc.use_color) : tinrc.use_color;
 				break;
 			}
 #endif /* HAVE_COLOR */
@@ -828,8 +877,10 @@ read_config_file(
 			break;
 		}
 	}
-	if (upgrade == RC_UPGRADE)
-		rc_post_update(fp);
+	if (!global_file && upgrade && upgrade->state == RC_UPGRADE)
+		rc_post_update(fp/* , upgrade */);
+
+	FreeAndNull(upgrade);
 	fclose(fp);
 
 	/*
@@ -853,10 +904,7 @@ read_config_file(
 		}
 	}
 	if (is_7bit) {
-		if (tinrc.mail_mime_encoding != MIME_ENCODING_7BIT)
-			tinrc.mail_mime_encoding = MIME_ENCODING_7BIT;
-		if (tinrc.post_mime_encoding != MIME_ENCODING_7BIT)
-			tinrc.post_mime_encoding = MIME_ENCODING_7BIT;
+		tinrc.mail_mime_encoding = tinrc.post_mime_encoding = MIME_ENCODING_7BIT;
 	} else {
 		if (tinrc.mail_mime_encoding == MIME_ENCODING_7BIT)
 			tinrc.mail_mime_encoding = MIME_ENCODING_QP;
@@ -873,7 +921,7 @@ read_config_file(
 
 	/* set defaults if blank */
 	if (!*tinrc.editor_format)
-		STRCPY(tinrc.editor_format, TIN_EDITOR_FMT_ON);
+		STRCPY(tinrc.editor_format, TIN_EDITOR_FMT);
 	if (!*tinrc.select_format)
 		STRCPY(tinrc.select_format, DEFAULT_SELECT_FORMAT);
 	if (!*tinrc.group_format)
@@ -882,10 +930,15 @@ read_config_file(
 		STRCPY(tinrc.thread_format, DEFAULT_THREAD_FORMAT);
 	if (!*tinrc.date_format)
 		STRCPY(tinrc.date_format, DEFAULT_DATE_FORMAT);
+	if (!*tinrc.inews_prog)
+		STRCPY(tinrc.inews_prog, INTERNAL_CMD);
 	/* determine local charset */
-#if defined(NO_LOCALE) && !defined(CHARSET_CONVERSION)
+#ifndef CHARSET_CONVERSION
+	if (!*tinrc.mm_charset)
+		STRCPY(tinrc.mm_charset, get_val("MM_CHARSET", MM_CHARSET));
 	strcpy(tinrc.mm_local_charset, tinrc.mm_charset);
-#endif /* NO_LOCALE && !CHARSET_CONVERSION */
+#endif /* !CHARSET_CONVERSION */
+
 	return TRUE;
 }
 
@@ -920,9 +973,6 @@ write_config_file(
 	fprintf(fp, "%s", _(txt_savedir.tinrc));
 	fprintf(fp, "savedir=%s\n\n", tinrc.savedir);
 
-	fprintf(fp, "%s", _(txt_auto_save.tinrc));
-	fprintf(fp, "auto_save=%s\n\n", print_boolean(tinrc.auto_save));
-
 	fprintf(fp, "%s", _(txt_mark_saved_read.tinrc));
 	fprintf(fp, "mark_saved_read=%s\n\n", print_boolean(tinrc.mark_saved_read));
 
@@ -939,7 +989,7 @@ write_config_file(
 	fprintf(fp, "prompt_followupto=%s\n\n", print_boolean(tinrc.prompt_followupto));
 
 	fprintf(fp, "%s", _(txt_confirm_choice.tinrc));
-	fprintf(fp, "confirm_choice=%s\n\n", txt_confirm_choices[tinrc.confirm_choice]);
+	fprintf(fp, "confirm_choice=%d\n\n", tinrc.confirm_choice);
 
 	fprintf(fp, "%s", _(txt_mark_ignore_tags.tinrc));
 	fprintf(fp, "mark_ignore_tags=%s\n\n", print_boolean(tinrc.mark_ignore_tags));
@@ -1027,9 +1077,6 @@ write_config_file(
 	fprintf(fp, "%s", _(txt_batch_save.tinrc));
 	fprintf(fp, "batch_save=%s\n\n", print_boolean(tinrc.batch_save));
 
-	fprintf(fp, "%s", _(txt_start_editor_offset.tinrc));
-	fprintf(fp, "start_editor_offset=%s\n\n", print_boolean(tinrc.start_editor_offset));
-
 	fprintf(fp, "%s", _(txt_editor_format.tinrc));
 	fprintf(fp, "editor_format=%s\n\n", tinrc.editor_format);
 
@@ -1122,6 +1169,11 @@ write_config_file(
 	fprintf(fp, "%s", _(txt_show_signatures.tinrc));
 	fprintf(fp, "show_signatures=%s\n\n", print_boolean(tinrc.show_signatures));
 
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+	fprintf(fp, "%s", _(txt_suppress_soft_hyphens.tinrc));
+	fprintf(fp, "suppress_soft_hyphens=%s\n\n", print_boolean(tinrc.suppress_soft_hyphens));
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
+
 	fprintf(fp, "%s", _(txt_tex2iso_conv.tinrc));
 	fprintf(fp, "tex2iso_conv=%s\n\n", print_boolean(tinrc.tex2iso_conv));
 
@@ -1142,31 +1194,31 @@ write_config_file(
 #endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 
 	fprintf(fp, "%s", _(txt_art_marked_deleted.tinrc));
-	fprintf(fp, "art_marked_deleted=%c\n\n", SPACE_TO_DASH(tinrc.art_marked_deleted));
+	fprintf(fp, "art_marked_deleted=%"T_CHAR_FMT"\n\n", SPACE_TO_DASH(tinrc.art_marked_deleted));
 
 	fprintf(fp, "%s", _(txt_art_marked_inrange.tinrc));
-	fprintf(fp, "art_marked_inrange=%c\n\n", SPACE_TO_DASH(tinrc.art_marked_inrange));
+	fprintf(fp, "art_marked_inrange=%"T_CHAR_FMT"\n\n", SPACE_TO_DASH(tinrc.art_marked_inrange));
 
 	fprintf(fp, "%s", _(txt_art_marked_return.tinrc));
-	fprintf(fp, "art_marked_return=%c\n\n", SPACE_TO_DASH(tinrc.art_marked_return));
+	fprintf(fp, "art_marked_return=%"T_CHAR_FMT"\n\n", SPACE_TO_DASH(tinrc.art_marked_return));
 
 	fprintf(fp, "%s", _(txt_art_marked_selected.tinrc));
-	fprintf(fp, "art_marked_selected=%c\n\n", SPACE_TO_DASH(tinrc.art_marked_selected));
+	fprintf(fp, "art_marked_selected=%"T_CHAR_FMT"\n\n", SPACE_TO_DASH(tinrc.art_marked_selected));
 
 	fprintf(fp, "%s", _(txt_art_marked_recent.tinrc));
-	fprintf(fp, "art_marked_recent=%c\n\n", SPACE_TO_DASH(tinrc.art_marked_recent));
+	fprintf(fp, "art_marked_recent=%"T_CHAR_FMT"\n\n", SPACE_TO_DASH(tinrc.art_marked_recent));
 
 	fprintf(fp, "%s", _(txt_art_marked_unread.tinrc));
-	fprintf(fp, "art_marked_unread=%c\n\n", SPACE_TO_DASH(tinrc.art_marked_unread));
+	fprintf(fp, "art_marked_unread=%"T_CHAR_FMT"\n\n", SPACE_TO_DASH(tinrc.art_marked_unread));
 
 	fprintf(fp, "%s", _(txt_art_marked_read.tinrc));
-	fprintf(fp, "art_marked_read=%c\n\n", SPACE_TO_DASH(tinrc.art_marked_read));
+	fprintf(fp, "art_marked_read=%"T_CHAR_FMT"\n\n", SPACE_TO_DASH(tinrc.art_marked_read));
 
 	fprintf(fp, "%s", _(txt_art_marked_killed.tinrc));
-	fprintf(fp, "art_marked_killed=%c\n\n", SPACE_TO_DASH(tinrc.art_marked_killed));
+	fprintf(fp, "art_marked_killed=%"T_CHAR_FMT"\n\n", SPACE_TO_DASH(tinrc.art_marked_killed));
 
 	fprintf(fp, "%s", _(txt_art_marked_read_selected.tinrc));
-	fprintf(fp, "art_marked_read_selected=%c\n\n", SPACE_TO_DASH(tinrc.art_marked_read_selected));
+	fprintf(fp, "art_marked_read_selected=%"T_CHAR_FMT"\n\n", SPACE_TO_DASH(tinrc.art_marked_read_selected));
 
 	fprintf(fp, "%s", _(txt_force_screen_redraw.tinrc));
 	fprintf(fp, "force_screen_redraw=%s\n\n", print_boolean(tinrc.force_screen_redraw));
@@ -1174,6 +1226,11 @@ write_config_file(
 	fprintf(fp, "%s", _(txt_inews_prog.tinrc));
 	fprintf(fp, "inews_prog=%s\n\n", tinrc.inews_prog);
 
+#ifdef USE_CANLOCK
+	fprintf(fp, "%s", _(txt_cancel_lock_algo.tinrc));
+	fprintf(fp, "cancel_lock_algo=%s\n\n", txt_cancel_lock_algos[tinrc.cancel_lock_algo]);
+#endif /* USE_CANLOCK */
+
 	fprintf(fp, "%s", _(txt_auto_list_thread.tinrc));
 	fprintf(fp, "auto_list_thread=%s\n\n", print_boolean(tinrc.auto_list_thread));
 
@@ -1266,10 +1323,8 @@ write_config_file(
 	fprintf(fp, "%s", _(txt_col_subject.tinrc));
 	fprintf(fp, "col_subject=%d\n\n", tinrc.col_subject);
 
-#ifdef HAVE_COLOR
 	fprintf(fp, "%s", _(txt_col_extquote.tinrc));
 	fprintf(fp, "col_extquote=%d\n\n", tinrc.col_extquote);
-#endif /* HAVE_COLOR */
 
 	fprintf(fp, "%s", _(txt_col_response.tinrc));
 	fprintf(fp, "col_response=%d\n\n", tinrc.col_response);
@@ -1388,6 +1443,9 @@ write_config_file(
 	fprintf(fp, "%s", _(txt_strip_bogus.tinrc));
 	fprintf(fp, "strip_bogus=%d\n\n", tinrc.strip_bogus);
 
+	fprintf(fp, "%s", _(txt_show_help_mail_sign.tinrc));
+	fprintf(fp, "show_help_mail_sign=%d\n\n", tinrc.show_help_mail_sign);
+
 	fprintf(fp, "%s", _(txt_select_format.tinrc));
 	fprintf(fp, "select_format=%s\n\n", tinrc.select_format);
 
@@ -1413,6 +1471,11 @@ write_config_file(
 	fprintf(fp, "render_bidi=%s\n\n", print_boolean(tinrc.render_bidi));
 #endif /* HAVE_LIBICUUC && MULTIBYTE_ABLE && HAVE_UNICODE_UBIDI_H && !NO_LOCALE */
 
+#ifdef NNTPS_ABLE
+	fprintf(fp, "%s", _(txt_tls_ca_cert_file.tinrc));
+	fprintf(fp, "tls_ca_cert_file=%s\n\n", tinrc.tls_ca_cert_file);
+#endif /* NNTPS_ABLE */
+
 	fprintf(fp, "%s", _(txt_tinrc_filter));
 	fprintf(fp, "default_filter_kill_header=%d\n", tinrc.default_filter_kill_header);
 	fprintf(fp, "default_filter_kill_global=%s\n", print_boolean(tinrc.default_filter_kill_global));
@@ -1467,7 +1530,13 @@ write_config_file(
 		}
 	}
 
+#ifdef HAVE_FCHMOD
 	fchmod(fileno(fp), (mode_t) (S_IRUSR|S_IWUSR)); /* rename_file() preserves mode */
+#else
+#	ifdef HAVE_CHMOD
+	chmod(file_tmp, (mode_t) (S_IRUSR|S_IWUSR)); /* rename_file() preserves mode */
+#	endif /* HAVE_CHMOD */
+#endif /* HAVE_FCHMOD */
 
 	if ((i = ferror(fp)) || fclose(fp)) {
 		error_message(2, _(txt_filesystem_full), CONFIG_FILE);
@@ -1621,15 +1690,11 @@ match_string(
 	char *dst,
 	size_t dstlen)
 {
-	char *ptr;
 	size_t patlen = strlen(pat);
 
-	if (STRNCMPEQ(line, pat, patlen) && (strlen(line) > patlen /* + 1 */)) {
-		if (dst != NULL && dstlen >= 1) {
-			strncpy(dst, &line[patlen], dstlen);
-			if ((ptr = strrchr(dst, '\n')) != NULL)
-				*ptr = '\0';
-		}
+	if (STRNCMPEQ(line, pat, patlen) && (strlen(line) > patlen)) {
+		if (dst != NULL && dstlen >= 1)
+			my_strncpy(dst, &line[patlen], dstlen - 1);
 		return TRUE;
 	}
 	return FALSE;
@@ -1648,13 +1713,14 @@ match_item(
 	char *nline = my_strdup(line);
 	size_t patlen = strlen(pat);
 
-	nline[strlen(nline) - 1] = '\0'; /* remove tailing \n */
+	if ((ptr = strchr(nline, '\n')) != NULL) /* terminate on \n */
+		*ptr = '\0';
 
 	if (!strcasecmp(nline, pat)) {
-		strncpy(dst, &nline[patlen], dstlen);
-		if ((ptr = strrchr(dst, '\n')) != NULL)
-			*ptr = '\0';
-
+		if (dst != NULL) {
+			strncpy(dst, &nline[patlen], dstlen);
+			dst[dstlen ? (dstlen - 1) : 0] = '\0';
+		}
 		free(nline);
 		return TRUE;
 	}
@@ -1751,7 +1817,7 @@ ulBuildArgv(
 				tmp++;
 			}
 			i++;
-			new_argv = my_realloc(new_argv, ((i + 1) * sizeof(char *)));
+			new_argv = my_realloc(new_argv, ((size_t) (i + 1) * sizeof(char *)));
 			new_argv[i] = NULL;
 		} else
 			tmp++;
@@ -1769,7 +1835,7 @@ static t_bool
 rc_update(
 	FILE *fp)
 {
-	char buf[1024];
+	char buf[LEN];
 	int show_info = 1;
 	t_bool auto_bcc = FALSE;
 	t_bool auto_cc = FALSE;
@@ -1802,7 +1868,7 @@ rc_update(
 		if (buf[0] == '#' || buf[0] == '\n')
 			continue;
 
-		switch (tolower((unsigned char) buf[0])) {
+		switch (my_tolower((unsigned char) buf[0])) {
 			case 'a':
 				if (match_boolean(buf, "auto_bcc=", &auto_bcc))
 					break;
@@ -1848,9 +1914,23 @@ rc_update(
 					}
 					break;
 				}
-				/* simple rename */
-				if (match_string(buf, "default_sigfile=", tinrc.sigfile, sizeof(tinrc.sigfile)))
+				/* 1. simple rename
+				 *
+				 * 2. previous versions has always passed groupname to external
+				 *    commands, now we look for %G
+				 */
+				if (match_string(buf, "default_sigfile=", tinrc.sigfile, sizeof(tinrc.sigfile))) {
+					size_t l = strlen(tinrc.sigfile);
+
+					if (tinrc.sigfile[0] == '!' && (tinrc.sigfile[l - 2] != '%' || tinrc.sigfile[l - 1] != 'G')) {
+						char *newbuf = my_malloc(sizeof(tinrc.sigfile) + 4);
+
+						sprintf(newbuf, "%s %s", tinrc.sigfile, "%G");
+						my_strncpy(tinrc.sigfile, newbuf, sizeof(tinrc.sigfile) - 1);
+						free(newbuf);
+					}
 					break;
+				}
 				break;
 
 			case 'h':
@@ -1888,7 +1968,7 @@ rc_update(
 					break;
 				if (match_boolean(buf, "show_last_line_prev_page=", &show_last_line_prev_page))
 					break;
-				if (match_boolean(buf, "show_lines=", &show_lines)){
+				if (match_boolean(buf, "show_lines=", &show_lines)) {
 					show_lines_or_score = TRUE;
 					break;
 				}
@@ -1943,7 +2023,7 @@ rc_update(
 		tinrc.hide_uue = 1;
 
 	if (keep_posted_articles)
-		strncpy(tinrc.posted_articles_file, "posted", sizeof(tinrc.posted_articles_file) - 1);
+		STRCPY(tinrc.posted_articles_file, "posted");
 
 	tinrc.quote_style = (compress_quotes ? QUOTE_COMPRESS : 0) + (quote_empty_lines ? QUOTE_EMPTY : 0) + (quote_signatures ? QUOTE_SIGS : 0);
 
@@ -1976,13 +2056,13 @@ rc_update(
 		tinrc.scroll_lines = -1;
 
 	if (use_builtin_inews)
-		strncpy(tinrc.inews_prog, INTERNAL_CMD, sizeof(tinrc.inews_prog) - 1);
+		STRCPY(tinrc.inews_prog, INTERNAL_CMD);
 
 	if (use_mailreader_i)
 		tinrc.interactive_mailer = INTERACTIVE_WITHOUT_HEADERS;
 
 	if (!use_metamail || getenv("NOMETAMAIL") != NULL)
-		strncpy(tinrc.metamail_prog, INTERNAL_CMD, sizeof(tinrc.metamail_prog) - 1);
+		STRCPY(tinrc.metamail_prog, INTERNAL_CMD);
 	else
 		my_strncpy(tinrc.metamail_prog, METAMAIL_CMD, sizeof(tinrc.metamail_prog) - 1);
 
@@ -1995,12 +2075,16 @@ rc_update(
  * auto update tinrc
  * called at the end of read_config_file()
  * useful to update variables which are already present in tinrc
+ *
+ * pass upgrade to this function once we need to exactly check
+ * upgrade->file_version, upgrade->current_version
  */
 static t_bool
 rc_post_update(
-	FILE *fp)
+	FILE *fp
+	/* , struct t_version *upgrade */)
 {
-	char buf[1024];
+	char buf[LEN];
 	int groupname_max_length = 0;
 
 	if (!fp)
@@ -2011,13 +2095,46 @@ rc_post_update(
 		if (buf[0] == '#' || buf[0] == '\n')
 			continue;
 
-		switch (tolower((unsigned char) buf[0])) {
+		switch (my_tolower((unsigned char) buf[0])) {
+			case 'c':
+#ifdef USE_CANLOCK
+				{
+					t_bool cancel_locks = TRUE;
+
+					if (match_boolean(buf, "cancel_locks=", &cancel_locks)) {
+						if (!cancel_locks)
+							tinrc.cancel_lock_algo = 0;
+						break;
+					}
+				}
+#endif /* USE_CANLOCK */
+				break;
+
 			case 'g':
 				if (match_integer(buf, "groupname_max_length=", &groupname_max_length, 132))
 					break;
 
 				break;
 
+			case 's':
+				/*
+				 * previous versions has always passed groupname to external
+				 * commands, now we look for %G
+				 */
+				if (match_string(buf, "sigfile=", tinrc.sigfile, sizeof(tinrc.sigfile))) {
+					size_t l = strlen(tinrc.sigfile);
+
+					if (tinrc.sigfile[0] == '!' && (tinrc.sigfile[l - 2] != '%' || tinrc.sigfile[l - 1] != 'G')) {
+						char *newbuf = my_malloc(sizeof(tinrc.sigfile) + 4);
+
+						sprintf(newbuf, "%s %s", tinrc.sigfile, "%G");
+						my_strncpy(tinrc.sigfile, newbuf, sizeof(tinrc.sigfile) - 1);
+						free(newbuf);
+					}
+					break;
+				}
+				break;
+
 			default:
 				break;
 		}
@@ -2059,7 +2176,7 @@ read_server_config(
 	char file[PATH_LEN];
 	char newnews_info[LEN];
 	char serverdir[PATH_LEN];
-	enum rc_state upgrade = RC_CHECK;
+	struct t_version *upgrade = NULL;
 
 #ifdef NNTP_ABLE
 	if (read_news_via_nntp && !read_saved_news && nntp_tcp_port != IPPORT_NNTP)
@@ -2072,10 +2189,12 @@ read_server_config(
 	joinpath(serverdir, sizeof(serverdir), rcdir, file);
 	joinpath(file, sizeof(file), serverdir, SERVERCONFIG_FILE);
 	joinpath(local_newsgroups_file, sizeof(local_newsgroups_file), serverdir, NEWSGROUPS_FILE);
+
 	if ((fp = fopen(file, "r")) == NULL)
 		return;
-	while (NULL != (line = tin_fgets(fp, FALSE))) {
-		if (('#' == *line) || ('\0' == *line))
+
+	while ((line = tin_fgets(fp, FALSE)) != NULL) {
+		if ((*line == '#') || (*line == '\0'))
 			continue;
 
 		if (match_string(line, "last_newnews=", newnews_info, sizeof(newnews_info))) {
@@ -2087,13 +2206,13 @@ read_server_config(
 			free(tmp_info);
 			continue;
 		}
+
 		if (match_string(line, "version=", NULL, 0)) {
-			if (RC_CHECK != upgrade)
-				/* ignore duplicate version lines; last match counts */
+			if (upgrade != NULL) /* ignore duplicate version lines; first match counts */
 				continue;
+
 			upgrade = check_upgrade(line, "version=", SERVERCONFIG_VERSION);
-			if (RC_IGNORE == upgrade)
-				/* Expected version number; nothing to do -> continue */
+			if (upgrade->state == RC_IGNORE) /* Expected version number; nothing to do -> continue */
 				continue;
 
 			/* Nothing to do yet for RC_UPGRADE and RC_DOWNGRADE */
@@ -2101,6 +2220,7 @@ read_server_config(
 		}
 	}
 	fclose(fp);
+	FreeAndNull(upgrade);
 }
 
 
@@ -2133,8 +2253,8 @@ write_server_config(
 	if ((no_write || post_article_and_exit || post_postponed_and_exit) && file_size(file) != -1L)
 		return;
 
-	if (-1 == stat(serverdir, &statbuf)) {
-		if (-1 == my_mkdir(serverdir, (mode_t) (S_IRWXU)))
+	if (stat(serverdir, &statbuf) == -1) {
+		if (my_mkdir(serverdir, (mode_t) (S_IRWXU)) == -1)
 			/* Can't create directory TODO: Add error handling */
 			return;
 	}
@@ -2143,7 +2263,7 @@ write_server_config(
 	file_tmp = get_tmpfilename(file);
 
 	if ((fp = fopen(file_tmp, "w")) == NULL) {
-		error_message(2, _(txt_filesystem_full_backup), SERVERCONFIG_FILE);
+		error_message(2, _(txt_filesystem_full_backup), SERVERCONFIG_FILE); /* TODO: better error handling/message */
 		free(file_tmp);
 		return;
 	}
@@ -2151,11 +2271,18 @@ write_server_config(
 	fprintf(fp, _(txt_serverconfig_header), PRODUCT, tin_progname, VERSION, RELEASEDATE, RELEASENAME, PRODUCT, PRODUCT);
 	fprintf(fp, "version=%s\n", SERVERCONFIG_VERSION);
 
-	if ((i = find_newnews_index(nntp_server)) >= 0)
+	if ((i = find_newnews_index(nntp_server)) >= 0) {
 		if (my_strftime(timestring, sizeof(timestring) - 1, "%Y-%m-%d %H:%M:%S UTC", gmtime(&(newnews[i].time))))
 			fprintf(fp, "last_newnews=%lu (%s)\n", (unsigned long int) newnews[i].time, timestring);
+	}
 
+#ifdef HAVE_FCHMOD
 	fchmod(fileno(fp), (mode_t) (S_IRUSR|S_IWUSR)); /* rename_file() preserves mode */
+#else
+#	ifdef HAVE_CHMOD
+	chmod(file_tmp, (mode_t) (S_IRUSR|S_IWUSR)); /* rename_file() preserves mode */
+#	endif /* HAVE_CHMOD */
+#endif /* HAVE_FCHMOD */
 
 	if ((i = ferror(fp)) || fclose(fp)) {
 		error_message(2, _(txt_filesystem_full), SERVERCONFIG_FILE);
