@@ -3,35 +3,38 @@
  *  Module    : filter.c
  *  Author    : I. Lea
  *  Created   : 1992-12-28
- *  Updated   : 2016-05-23
+ *  Updated   : 2022-08-29
  *  Notes     : Filter articles. Kill & auto selection are supported.
  *
- * Copyright (c) 1991-2016 Iain Lea <iain@bricbrac.de>
+ * Copyright (c) 1991-2023 Iain Lea <iain@bricbrac.de>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
- * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer.
+ *
+ * 1. Redistributions of source code must retain the above copyright notice,
+ *    this list of conditions and the following disclaimer.
+ *
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
- * 3. The name of the author may not be used to endorse or promote
- *    products derived from this software without specific prior written
- *    permission.
  *
- * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
- * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
- * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
- * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
- * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
- * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
- * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
- * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
- * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ * 3. Neither the name of the copyright holder nor the names of its
+ *    contributors may be used to endorse or promote products derived from
+ *    this software without specific prior written permission.
+ *
+ * * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+ * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
+ * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
+ * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
+ * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
+ * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
+ * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
+ * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
+ * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+ * POSSIBILITY OF SUCH DAMAGE.
  */
 
 
@@ -99,7 +102,7 @@ static struct t_filter_comment *add_filter_comment(struct t_filter_comment *ptr,
 static struct t_filter_comment *free_filter_comment(struct t_filter_comment *ptr);
 static struct t_filter_comment *copy_filter_comment(struct t_filter_comment *from, struct t_filter_comment *to);
 static t_bool add_filter_rule(struct t_group *group, struct t_article *art, struct t_filter_rule *rule, t_bool quick_filter_rule);
-static int test_regex(const char *string, char *regex, t_bool nocase, struct regex_cache *cache);
+static int test_match(const char *string, char *regex, t_bool nocase, t_bool use_regex, struct regex_cache *cache);
 static void expand_filter_array(struct t_filters *ptr);
 static void fmt_filter_menu_prompt(char *dest, size_t dest_len, const char *fmt_str, int len, const char *text);
 static void free_filter_item(struct t_filter *ptr);
@@ -180,7 +183,7 @@ expand_filter_array(
 
 	num = ++ptr->max;
 
-	block = num * sizeof(struct t_filter);
+	block = (size_t) num * sizeof(struct t_filter);
 
 	if (num == 1)	/* allocate */
 		ptr->filter = my_malloc(block);
@@ -197,29 +200,30 @@ expand_filter_array(
  * In case of error prints an error message and returns -1
  */
 static int
-test_regex(
+test_match(
 	const char *string,
 	char *regex,
 	t_bool nocase,
+	t_bool use_regex,
 	struct regex_cache *cache)
 {
-	int regex_errpos;
+	int error;
 
-	if (!tinrc.wildcard) {
+	if (!use_regex) {
 		if (wildmat(string, regex, nocase))
 			return 1;
 	} else {
 		if (!cache->re)
-			compile_regex(regex, cache, (nocase ? PCRE_CASELESS : 0));
+			compile_regex(regex, cache, (nocase ? REGEX_CASELESS : 0));
 		if (cache->re) {
-			regex_errpos = pcre_exec(cache->re, cache->extra, string, strlen(string), 0, 0, NULL, 0);
-			if (regex_errpos >= 0)
+			error = match_regex_ex(string, (int) strlen(string), 0, 0, cache);
+			if (error >= 0)
 				return 1;
-			else if (regex_errpos != PCRE_ERROR_NOMATCH) { /* also exclude PCRE_ERROR_BADUTF8 ? */
-				error_message(2, _(txt_pcre_error_num), regex_errpos);
+			else if (error != REGEX_ERROR_NOMATCH) { /* also exclude BADUTF8 ? */
+				error_message(2, _(txt_pcre_error_num), error);
 #ifdef DEBUG
 				if (debug & DEBUG_FILTER) {
-					debug_print_file("FILTER", _(txt_pcre_error_num), regex_errpos);
+					debug_print_file("FILTER", _(txt_pcre_error_num), error);
 					debug_print_file("FILTER", "\t regex: %s", regex);
 					debug_print_file("FILTER", "\tstring: %s", string);
 				}
@@ -254,6 +258,7 @@ set_filter(
 		ptr->gnksa_num = 0;
 		ptr->score = 0;
 		ptr->xref = NULL;
+		ptr->path = NULL;
 		ptr->time = (time_t) 0;
 		ptr->next = (struct t_filter *) 0;
 	}
@@ -273,6 +278,7 @@ free_filter_item(
 	FreeAndNull(ptr->from);
 	FreeAndNull(ptr->msgid);
 	FreeAndNull(ptr->xref);
+	FreeAndNull(ptr->path);
 }
 
 
@@ -313,6 +319,7 @@ read_filter_file(
 	char buffer[HEADER_LEN];
 	char gnksa[HEADER_LEN];
 	char xref[HEADER_LEN];
+	char path[HEADER_LEN];
 	char scbuf[PATH_LEN];
 	int i = 0;
 	int icase = 0;
@@ -320,11 +327,12 @@ read_filter_file(
 	long secs = 0L;
 	struct t_filter_comment *comment = NULL;
 	struct t_filter *ptr = NULL;
-	t_bool expired = FALSE;
+	t_bool need_write = FALSE;
+	t_bool no_version_line = TRUE;
 	t_bool expired_time = FALSE;
 	time_t current_secs = (time_t) 0;
 	static t_bool first_read = TRUE;
-	enum rc_state upgrade = RC_CHECK;
+	struct t_version *upgrade = NULL;
 
 	if ((fp = fopen(file, "r")) == NULL)
 		return FALSE;
@@ -348,16 +356,17 @@ read_filter_file(
 		if (*buf == '#') {
 			if (scope[0] == '\0')
 				filter_file_offset++;
-			if (upgrade == RC_CHECK && first_read && match_string(buf, "# Filter file V", NULL, 0)) {
+			if (upgrade == NULL && first_read && match_string(buf, "# Filter file V", NULL, 0)) {
 				first_read = FALSE;
+				no_version_line = FALSE;
 				upgrade = check_upgrade(buf, "# Filter file V", FILTER_VERSION);
-				if (upgrade != RC_IGNORE)
-					upgrade_prompt_quit(upgrade, FILTER_FILE); /* TODO: do something (more) useful here */
+				if (upgrade->state != RC_IGNORE)
+					upgrade_prompt_quit(upgrade, file); /* FILTER_FILE */ /* TODO: do something (more) useful here */
 			}
 			continue;
 		}
 
-		switch (tolower((unsigned char) buf[0])) {
+		switch (my_tolower((unsigned char) buf[0])) {
 			case 'c':
 				if (match_integer(buf + 1, "ase=", &icase, 1)) {
 					if (ptr && !expired_time)
@@ -411,6 +420,7 @@ read_filter_file(
 					msgid[0] = '\0';
 					buffer[0] = '\0';
 					xref[0] = '\0';
+					path[0] = '\0';
 					icase = 0;
 					secs = 0L;
 					break;
@@ -495,6 +505,23 @@ read_filter_file(
 				}
 				break;
 
+			case 'p':
+				if (match_string(buf + 1, "ath=", path, sizeof(path))) {
+					str_trim(path);
+					if (ptr && !expired_time) {
+						if (tinrc.wildcard && ptr[i].path != NULL) {
+							/* merge with already read value */
+							ptr[i].path = my_realloc(ptr[i].path, strlen(ptr[i].path) + strlen(path) + 2);
+							strcat(ptr[i].path, "|");
+							strcat(ptr[i].path, path);
+						} else {
+							FreeIfNeeded(ptr[i].path);
+							ptr[i].path = my_strdup(path);
+						}
+					}
+				}
+				break;
+
 			case 'r':
 				if (match_string(buf + 1, "efs_only=", msgid, sizeof(msgid))) {
 					if (ptr && !expired_time) {
@@ -575,7 +602,7 @@ read_filter_file(
 #endif /* DEBUG */
 							glob_filter.num--;
 							expired_time = TRUE;
-							expired = TRUE;
+							need_write = TRUE;
 						}
 					}
 				}
@@ -583,12 +610,12 @@ read_filter_file(
 
 			case 'x':
 				/*
-				 * TODO: fromat has changed in FILTER_VERSION 1.0.0,
+				 * TODO: format has changed in FILTER_VERSION 1.0.0,
 				 *       should we comment out older xref rules like below?
 				 */
-				if (match_string(buf + 1, "ref=", xref, sizeof(xref))) {
+				if (ptr && match_string(buf + 1, "ref=", xref, sizeof(xref))) {
 					str_trim(xref);
-					if (ptr && !expired_time) {
+					if (!expired_time) {
 						if (tinrc.wildcard && ptr[i].xref != NULL) {
 							/* merge with already read value */
 							ptr[i].xref = my_realloc(ptr[i].xref, strlen(ptr[i].xref) + strlen(xref) + 2);
@@ -601,23 +628,21 @@ read_filter_file(
 					}
 					break;
 				}
-				if (upgrade == RC_UPGRADE) {
+				if (ptr && ((upgrade && upgrade->state == RC_UPGRADE) || no_version_line)) {
 					char foo[HEADER_LEN];
 
 					if (match_string(buf + 1, "ref_max=", foo, LEN - 1)) {
-						/*
-						 * TODO: add to the right rule, give better explanation.
-						 */
+						/* TODO: give better explanation. */
 						snprintf(foo, HEADER_LEN, "%s%s", _(txt_removed_rule), str_trim(buf));
-						comment = add_filter_comment(comment, foo);
+						ptr[i].comment = add_filter_comment(ptr[i].comment, foo);
+						need_write = TRUE;
 						break;
 					}
 					if (match_string(buf + 1, "ref_score=", foo, LEN - 1)) {
-						/*
-						 * TODO: add to the right rule, give better explanation.
-						 */
+						/* TODO: give better explanation. */
 						snprintf(foo, HEADER_LEN, "%s%s", _(txt_removed_rule), str_trim(buf));
-						comment = add_filter_comment(comment, foo);
+						ptr[i].comment = add_filter_comment(ptr[i].comment, foo);
+						need_write = TRUE;
 					}
 				}
 				break;
@@ -626,14 +651,26 @@ read_filter_file(
 				break;
 		}
 	}
+
+	if (comment) /* stray comment without scope */
+		(void) free_filter_comment(comment);
+
 	fclose(fp);
 
-	if (expired || upgrade == RC_UPGRADE)
+	if (!upgrade && need_write) {
+		upgrade = my_malloc(sizeof(struct t_version));
+		upgrade->state = RC_UPGRADE;
+		upgrade->file_version = -1;
+		upgrade_prompt_quit(upgrade, file); /* TODO: do something (more) useful here */
+	}
+
+	if (need_write || (upgrade && upgrade->state == RC_UPGRADE))
 		write_filter_file(file);
 
 	if (!cmd_line && !batch_mode)
 		clear_message();
 
+	FreeAndNull(upgrade);
 	return TRUE;
 }
 
@@ -840,6 +877,9 @@ write_filter_array(
 		if (ptr->filter[i].xref != NULL)
 			fprintf(fp, "xref=%s\n", ptr->filter[i].xref);
 
+		if (ptr->filter[i].path != NULL)
+			fprintf(fp, "path=%s\n", ptr->filter[i].path);
+
 		if (ptr->filter[i].time) {
 			char timestring[25];
 			if (my_strftime(timestring, sizeof(timestring) - 1, "%Y-%m-%d %H:%M:%S UTC", gmtime(&(ptr->filter[i].time))))
@@ -884,9 +924,9 @@ get_choice(
 				break;
 
 			case ESC:	/* (ESC) common arrow keys */
-#	ifdef HAVE_KEY_PREFIX
+#ifdef HAVE_KEY_PREFIX
 			case KEY_PREFIX:
-#	endif /* HAVE_KEY_PREFIX */
+#endif /* HAVE_KEY_PREFIX */
 				switch (get_arrow_key(ch)) {
 					case KEYMAP_UP:
 						i--;
@@ -909,7 +949,7 @@ get_choice(
 			default:
 				break;
 		}
-	} while (ch != '\n' && ch != '\r' && ch != iKeyAbort); /* TODO: replace hardcoded keynames */
+	} while (ch != '\n' && ch != '\r' && ch != iKeyAbort); /* TODO: replace hard coded keynames */
 
 	if (ch == iKeyAbort)
 		return -1;
@@ -998,7 +1038,7 @@ fmt_filter_menu_prompt(
 		if ((buf = wchar_t2char(wbuf2)) == NULL) {
 			/* conversion failed, truncate original string */
 			buf = my_malloc(len + 1);
-			snprintf(buf, len + 1, "%-*.*s", len, len, text);
+			snprintf(buf, (size_t) (len + 1), "%-*.*s", len, len, text);
 		}
 
 		free(wbuf);
@@ -1007,12 +1047,10 @@ fmt_filter_menu_prompt(
 #endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 	{
 		buf = my_malloc(len + 1);
-		snprintf(buf, len + 1, "%-*.*s", len, len, text);
+		snprintf(buf, (size_t) (len + 1), "%-*.*s", len, len, text);
 	}
 	snprintf(dest, dest_len, fmt_str, buf);
 	free(buf);
-
-	return;
 }
 
 
@@ -1068,9 +1106,9 @@ filter_menu(
 	/*
 	 * setup correct text for user selected menu
 	 */
-	printascii(keyedit, func_to_key(FILTER_EDIT, filter_keys));
-	printascii(keyquit, func_to_key(GLOBAL_QUIT, filter_keys));
-	printascii(keysave, func_to_key(FILTER_SAVE, filter_keys));
+	PrintFuncKey(keyedit, FILTER_EDIT, filter_keys);
+	PrintFuncKey(keyquit, GLOBAL_QUIT, filter_keys);
+	PrintFuncKey(keysave, FILTER_SAVE, filter_keys);
 
 	if (type == GLOBAL_MENU_FILTER_KILL) {
 		ptr_filter_from = _(txt_kill_from);
@@ -1387,7 +1425,7 @@ filter_menu(
 			*(++ptr) = '*';
 			*(++ptr) = '\0';
 			j++;
-			list = my_realloc(list, sizeof(char *) * (j + 1)); /* one element more */
+			list = my_realloc(list, sizeof(char *) * (size_t) (j + 1)); /* one element more */
 			list[j] = my_strdup(list[j - 1]);
 			list[j][strlen(list[j]) - 2] = '\0';
 		}
@@ -1617,8 +1655,8 @@ add_filter_rule(
 	struct t_filter_rule *rule,
 	t_bool quick_filter_rule)
 {
-	char acBuf[PATH_LEN];
-	char sbuf[(sizeof(acBuf) / 2)]; /* half as big as acBuf so quote_wild(sbuf) fits into acBuf */
+	char acbuf[PATH_LEN];
+	char sbuf[(sizeof(acbuf) / 2)]; /* half as big as acbuf so quote_wild(sbuf) fits into acbuf */
 	int i = glob_filter.num;
 	t_bool filtered = FALSE;
 	time_t current_time;
@@ -1629,20 +1667,21 @@ add_filter_rule(
 
 	ptr = glob_filter.filter;
 
-	ptr[i].icase = FALSE;
 	ptr[i].inscope = TRUE;
+	ptr[i].icase = FALSE;
 	ptr[i].fullref = FILTER_MSGID;
 	ptr[i].comment = (struct t_filter_comment *) 0;
 	ptr[i].scope = NULL;
 	ptr[i].subj = NULL;
 	ptr[i].from = NULL;
 	ptr[i].msgid = NULL;
-	ptr[i].lines_cmp = rule->lines_cmp;
+	ptr[i].lines_cmp = (char) rule->lines_cmp;
 	ptr[i].lines_num = rule->lines_num;
 	ptr[i].gnksa_cmp = FILTER_LINES_NO;
 	ptr[i].gnksa_num = 0;
 	ptr[i].score = rule->score;
 	ptr[i].xref = NULL;
+	ptr[i].path = NULL;
 
 	if (rule->comment != NULL)
 		ptr[i].comment = copy_filter_comment(rule->comment, ptr[i].comment);
@@ -1675,24 +1714,24 @@ add_filter_rule(
 
 	ptr[i].icase = rule->icase;
 	if (*rule->text) {
-		snprintf(acBuf, sizeof(acBuf), REGEX_FMT, quote_wild_whitespace(rule->text));
+		snprintf(acbuf, sizeof(acbuf), REGEX_FMT, quote_wild_whitespace(rule->text));
 
 		switch (rule->counter) {
 			case FILTER_SUBJ_CASE_IGNORE:
 			case FILTER_SUBJ_CASE_SENSITIVE:
-				ptr[i].subj = my_strdup(acBuf);
+				ptr[i].subj = my_strdup(acbuf);
 				break;
 
 			case FILTER_FROM_CASE_IGNORE:
 			case FILTER_FROM_CASE_SENSITIVE:
-				ptr[i].from = my_strdup(acBuf);
+				ptr[i].from = my_strdup(acbuf);
 				break;
 
 			case FILTER_MSGID:
 			case FILTER_MSGID_LAST:
 			case FILTER_MSGID_ONLY:
 			case FILTER_REFS_ONLY:
-				ptr[i].msgid = my_strdup(acBuf);
+				ptr[i].msgid = my_strdup(acbuf);
 				ptr[i].fullref = rule->counter;
 				break;
 
@@ -1706,17 +1745,17 @@ add_filter_rule(
 	} else {
 		/*
 		 * STRCPY() truncates subject/from/message-id so it fits
-		 * into acBuf even after quote_wild()
+		 * into acbuf even after quote_wild()
 		 */
 		if (rule->subj_ok) {
 			STRCPY(sbuf, art->subject);
-			snprintf(acBuf, sizeof(acBuf), REGEX_FMT, (rule->check_string ? quote_wild(sbuf) : sbuf));
-			ptr[i].subj = my_strdup(acBuf);
+			snprintf(acbuf, sizeof(acbuf), REGEX_FMT, (rule->check_string ? quote_wild(sbuf) : sbuf));
+			ptr[i].subj = my_strdup(acbuf);
 		}
 		if (rule->from_ok) {
 			STRCPY(sbuf, art->from);
-			snprintf(acBuf, sizeof(acBuf), REGEX_FMT, quote_wild(sbuf));
-			ptr[i].from = my_strdup(acBuf);
+			snprintf(acbuf, sizeof(acbuf), REGEX_FMT, quote_wild(sbuf));
+			ptr[i].from = my_strdup(acbuf);
 		}
 		/*
 		 * message-ids should be quoted
@@ -1749,8 +1788,8 @@ add_filter_rule(
 			} else {
 				STRCPY(sbuf, MSGID(art));
 			}
-			snprintf(acBuf, sizeof(acBuf), REGEX_FMT, quote_wild(sbuf));
-			ptr[i].msgid = my_strdup(acBuf);
+			snprintf(acbuf, sizeof(acbuf), REGEX_FMT, quote_wild(sbuf));
+			ptr[i].msgid = my_strdup(acbuf);
 			ptr[i].fullref = rule->fullref;
 		}
 		if (rule->subj_ok || rule->from_ok || rule->msgid_ok || rule->lines_ok) {
@@ -1762,7 +1801,7 @@ add_filter_rule(
 	if (filtered) {
 #ifdef DEBUG
 		if (debug & DEBUG_FILTER)
-			wait_message(2, "inscope=[%s] scope=[%s] case=[%d] subj=[%s] from=[%s] msgid=[%s] fullref=[%d] line=[%d %d] time=[%lu]", bool_unparse(ptr[i].inscope), BlankIfNull(rule->scope), ptr[i].icase, BlankIfNull(ptr[i].subj), BlankIfNull(ptr[i].from), BlankIfNull(ptr[i].msgid), ptr[i].fullref, ptr[i].lines_cmp, ptr[i].lines_num, (unsigned long int) ptr[i].time);
+			wait_message(2, "inscope=[%s] scope=[%s] case=[%d] subj=[%s] from=[%s] msgid=[%s] fullref=[%u] line=[%d %d] time=[%lu]", bool_unparse(ptr[i].inscope), BlankIfNull(rule->scope), ptr[i].icase, BlankIfNull(ptr[i].subj), BlankIfNull(ptr[i].from), BlankIfNull(ptr[i].msgid), ptr[i].fullref, ptr[i].lines_cmp, ptr[i].lines_num, (unsigned long int) ptr[i].time);
 #endif /* DEBUG */
 		write_filter_file(filter_file);
 	}
@@ -1799,7 +1838,7 @@ unfilter_articles(
  * Filter any articles in specified group.
  * Apply global filter rules followed by group filter rules.
  * In global rules check if scope field set to determine if
- * filter applys to current group.
+ * filter applies to current group.
  */
 t_bool
 filter_articles(
@@ -1813,8 +1852,10 @@ filter_articles(
 	struct regex_cache *regex_cache_from = NULL;
 	struct regex_cache *regex_cache_msgid = NULL;
 	struct regex_cache *regex_cache_xref = NULL;
+	struct regex_cache *regex_cache_path = NULL;
 	t_bool filtered = FALSE;
 	t_bool error = FALSE;
+	t_bool use_regex = tinrc.wildcard;
 
 	/*
 	 * check if there are any global filter rules
@@ -1839,32 +1880,33 @@ filter_articles(
 	 * set up cache tables for all types of filter rules
 	 * (only for regexp matching)
 	 */
-	if (tinrc.wildcard) {
+	if (use_regex) {
 		size_t msiz;
 
-		msiz = sizeof(struct regex_cache) * num;
+		msiz = sizeof(struct regex_cache) * (size_t) num;
 		regex_cache_subj = my_malloc(msiz);
 		regex_cache_from = my_malloc(msiz);
 		regex_cache_msgid = my_malloc(msiz);
 		regex_cache_xref = my_malloc(msiz);
+		regex_cache_path = my_malloc(msiz);
 		for (j = 0; j < num; j++) {
-			regex_cache_subj[j].re = NULL;
-			regex_cache_subj[j].extra = NULL;
-			regex_cache_from[j].re = NULL;
-			regex_cache_from[j].extra = NULL;
-			regex_cache_msgid[j].re = NULL;
-			regex_cache_msgid[j].extra = NULL;
-			regex_cache_xref[j].re = NULL;
-			regex_cache_xref[j].extra = NULL;
+			regex_cache_init(&regex_cache_subj[j]);
+			regex_cache_init(&regex_cache_from[j]);
+			regex_cache_init(&regex_cache_msgid[j]);
+			regex_cache_init(&regex_cache_xref[j]);
+			regex_cache_init(&regex_cache_path[j]);
 		}
 	}
 
 	/*
 	 * loop through all arts applying global & local filtering rules
-	 *
-	 * TODO: allow iKeyAbort to stop filtering
 	 */
-	for (i = 0; (i < top_art) && !error; i++) {
+	for (i = 0; i < top_art && !error; i++) {
+#ifdef HAVE_SELECT
+		if (wait_for_input()) /* allow abort */
+			break; /* to free the mem */
+#endif /* HAVE_SELECT */
+
 		arts[i].score = 0;
 
 		if (tinrc.kill_level == KILL_UNREAD && IS_READ(i)) /* skip only when the article is read */
@@ -1876,7 +1918,7 @@ filter_articles(
 				 * Filter on Subject: line
 				 */
 				if (ptr[j].subj != NULL) {
-					switch (test_regex(arts[i].subject, ptr[j].subj, ptr[j].icase, &regex_cache_subj[j])) {
+					switch (test_match(arts[i].subject, ptr[j].subj, ptr[j].icase, use_regex, &regex_cache_subj[j])) {
 						case 1:
 							SET_FILTER(group, i, j);
 							break;
@@ -1899,7 +1941,7 @@ filter_articles(
 					else
 						STRCPY(buf, arts[i].from);
 
-					switch (test_regex(buf, ptr[j].from, ptr[j].icase, &regex_cache_from[j])) {
+					switch (test_match(buf, ptr[j].from, ptr[j].icase, use_regex, &regex_cache_from[j])) {
 						case 1:
 							SET_FILTER(group, i, j);
 							break;
@@ -1962,8 +2004,8 @@ filter_articles(
 							break;
 					}
 
-					if ((x = test_regex(myrefs, ptr[j].msgid, FALSE, &regex_cache_msgid[j])) == 0) /* no match */
-						x = test_regex(mymsgid, ptr[j].msgid, FALSE, &regex_cache_msgid[j]);
+					if ((x = test_match(myrefs, ptr[j].msgid, FALSE, use_regex, &regex_cache_msgid[j])) == 0) /* no match */
+						x = test_match(mymsgid, ptr[j].msgid, FALSE, use_regex, &regex_cache_msgid[j]);
 
 					switch (x) {
 						case 1:
@@ -2066,7 +2108,7 @@ filter_articles(
 							}
 						}
 #endif /* DEBUG */
-						if (strlen(s)) {
+						if (*s) {
 							/* reformat */
 							k = e = my_malloc(strlen(s) + 1);
 							while (*s) {
@@ -2084,51 +2126,71 @@ filter_articles(
 						} else {
 #ifdef DEBUG
 							if (debug & DEBUG_FILTER) /* TODO: lang.c, _()? */
-								debug_print_file("FILTER", "Skipping xref filter");
+								debug_print_file("FILTER", "Skipping Xref filter");
 #endif /* DEBUG */
 							error = TRUE;
 							break;
 						}
 
-						if (ptr[j].xref != NULL) {
-							switch (test_regex(k, ptr[j].xref, ptr[j].icase, &regex_cache_xref[j])) {
-								case 1:
-									SET_FILTER(group, i, j);
-									break;
+						switch (test_match(k, ptr[j].xref, ptr[j].icase, use_regex, &regex_cache_xref[j])) {
+							case 1:
+								SET_FILTER(group, i, j);
+								break;
 
-								case -1:
-									error = TRUE;
-									break;
+							case -1:
+								error = TRUE;
+								break;
 
-								default:
-									break;
-							}
+							default:
+								break;
 						}
 						free(k);
 					}
 				}
+
+				/*
+				 * Filter on Path: lines
+				 */
+				if (arts[i].path && *arts[i].path) {
+					if (ptr[j].path != NULL) {
+						switch (test_match(arts[i].path, ptr[j].path, ptr[j].icase, use_regex, &regex_cache_path[j])) {
+							case 1:
+								SET_FILTER(group, i, j);
+								break;
+
+							case -1:
+								error = TRUE;
+								break;
+
+							default:
+								break;
+						}
+					}
+				}
 			}
 		}
+		/*
+		 * if (i % (MODULO_COUNT_NUM * 20) == 0)
+		 *    show_progress("Filter", i, top_art);
+		 */
 	}
 
 	/*
 	 * throw away the contents of all regex_caches
 	 */
-	if (tinrc.wildcard) {
+	if (use_regex) {
 		for (j = 0; j < num; j++) {
-			FreeIfNeeded(regex_cache_subj[j].re);
-			FreeIfNeeded(regex_cache_subj[j].extra);
-			FreeIfNeeded(regex_cache_from[j].re);
-			FreeIfNeeded(regex_cache_from[j].extra);
-			FreeIfNeeded(regex_cache_msgid[j].re);
-			FreeIfNeeded(regex_cache_msgid[j].extra);
-			FreeIfNeeded(regex_cache_xref[j].re);
-			FreeIfNeeded(regex_cache_xref[j].extra);
+			regex_cache_destroy(&regex_cache_subj[j]);
+			regex_cache_destroy(&regex_cache_from[j]);
+			regex_cache_destroy(&regex_cache_msgid[j]);
+			regex_cache_destroy(&regex_cache_xref[j]);
+			regex_cache_destroy(&regex_cache_path[j]);
 		}
 		free(regex_cache_subj);
 		free(regex_cache_from);
 		free(regex_cache_msgid);
 		free(regex_cache_xref);
+		free(regex_cache_path);
 	}
 
 	/*
@@ -2151,6 +2213,7 @@ filter_articles(
 			}
 		}
 	}
+
 	if (!cmd_line && !batch_mode)
 		clear_message();
 
@@ -2158,6 +2221,33 @@ filter_articles(
 }
 
 
+/*
+ * Check if we have to filter on Path: for the given group
+ */
+t_bool
+filter_on_path(
+	struct t_group *group)
+{
+	int i;
+	struct t_filter *flt;
+	t_bool ret = FALSE;
+
+	if (group->glob_filter->num == 0)
+		return ret;
+
+	if (set_filter_scope(group)) {
+		flt = group->glob_filter->filter;
+		for (i = 0; i < group->glob_filter->num; i++) {
+			if (flt[i].inscope && flt[i].path) {
+				ret = TRUE;
+				break;
+			}
+		}
+	}
+	return ret;
+}
+
+
 static int
 set_filter_scope(
 	struct t_group *group)
@@ -2185,7 +2275,8 @@ set_filter_scope(
 
 
 /*
- * This will come in useful for filtering on non-overview hdr fields
+ * This will come in useful for filtering on non-overview hdr fields.
+ * See also art.c:get_path_header().
  */
 #if 0
 static FILE *
