@@ -3,35 +3,38 @@
  *  Module    : rfc2046.c
  *  Author    : Jason Faultless <jason@altarstone.com>
  *  Created   : 2000-02-18
- *  Updated   : 2016-02-11
+ *  Updated   : 2022-04-09
  *  Notes     : RFC 2046 MIME article parsing
  *
- * Copyright (c) 2000-2016 Jason Faultless <jason@altarstone.com>
+ * Copyright (c) 2000-2023 Jason Faultless <jason@altarstone.com>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
- * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer.
+ *
+ * 1. Redistributions of source code must retain the above copyright notice,
+ *    this list of conditions and the following disclaimer.
+ *
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
- * 3. The name of the author may not be used to endorse or promote
- *    products derived from this software without specific prior written
- *    permission.
  *
- * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
- * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
- * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
- * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
- * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
- * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
- * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
- * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
- * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ * 3. Neither the name of the copyright holder nor the names of its
+ *    contributors may be used to endorse or promote products derived from
+ *    this software without specific prior written permission.
+ *
+ * * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+ * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
+ * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
+ * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
+ * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
+ * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
+ * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
+ * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
+ * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+ * POSSIBILITY OF SUCH DAMAGE.
  */
 
 
@@ -54,7 +57,7 @@ static int count_lines(char *line);
 static int parse_multipart_article(FILE *infile, t_openartinfo *artinfo, t_part *part, int depth, t_bool show_progress_meter);
 static int parse_normal_article(FILE *in, t_openartinfo *artinfo, t_bool show_progress_meter);
 static int parse_rfc2045_article(FILE *infile, int line_count, t_openartinfo *artinfo, t_bool show_progress_meter);
-static unsigned int parse_content_encoding(const char *encoding);
+static unsigned int parse_content_encoding(char *encoding);
 static void decode_value(const char *charset, t_param *part);
 static void parse_content_type(char *type, t_part *content);
 static void parse_content_disposition(char *disp, t_part *part);
@@ -138,36 +141,60 @@ boundary_cmp(
 {
 	size_t blen = strlen(boundary);
 	size_t len;
+	char *e, *l;
 	int nl;
 
 	if ((len = strlen(line)) == 0)
 		return BOUND_NONE;
 
-	nl = line[len - 1] == '\n';
-
-	if (len != blen + 2 + nl && len != blen + 4 + nl)
+	if (blen + 2 > len)
 		return BOUND_NONE;
 
-	if (line[0] != '-' || line[1] != '-')
-		return BOUND_NONE;
+	/* remove trailing whites as per RFC 2046 5.1.1 */
+	l = my_strdup(line);
+	e = l + len - 1;
+	while (e > l + blen + 1 && isspace((unsigned char) *e))
+		*e-- = '\0';
 
-	if (strncmp(line + 2, boundary, blen) != 0)
-		return BOUND_NONE;
+	len = strlen(l);
 
-	if (line[blen + 2] != '-') {
-		if (nl ? line[blen + 2] == '\n' : line[blen + 2] == '\0')
-			return BOUND_START;
-		else
-			return BOUND_NONE;
+	nl = l[len - 1] == '\n';
+
+	if (len != blen + 2 + (size_t) nl && len != blen + 4 + (size_t) nl) {
+		free(l);
+		return BOUND_NONE;
+	}
+	if (l[0] != '-' || l[1] != '-') {
+		free(l);
+		return BOUND_NONE;
 	}
 
-	if (line[blen + 3] != '-')
+	if (strncmp(l + 2, boundary, blen) != 0) {
+		free(l);
 		return BOUND_NONE;
+	}
 
-	if (nl ? line[blen + 4] == '\n' : line[blen + 4] == '\0')
-		return BOUND_END;
-	else
+	if (l[blen + 2] != '-') {
+		if (nl ? l[blen + 2] == '\n' : l[blen + 2] == '\0') {
+			free(l);
+			return BOUND_START;
+		} else {
+			free(l);
+			return BOUND_NONE;
+		}
+	}
+
+	if (l[blen + 3] != '-') {
+		free(l);
 		return BOUND_NONE;
+	}
+
+	if (nl ? l[blen + 4] == '\n' : l[blen + 4] == '\0') {
+		free(l);
+		return BOUND_END;
+	}
+	free(l);
+	return BOUND_NONE;
 }
 
 
@@ -202,7 +229,7 @@ static char *
 skip_space(
 	char *source)
 {
-	while ((*source) && ((' ' == *source) || ('\t' == *source)))
+	while ((*source) && ((*source == ' ') || (*source == '\t')))
 		source++;
 	return *source ? source : NULL;
 }
@@ -232,6 +259,11 @@ remove_cwsp(
 		}
 
 		if (!inquotes) {
+			/* skip over quoted pairs */
+			if (c_cnt && src == '\\') {
+				++from;
+				continue;
+			}
 			if (src == '(') {
 				++c_cnt;
 				continue;
@@ -289,15 +321,15 @@ get_quoted_string(
 	ptr = *dest;
 	source++; /* skip over double quote */
 	while (*source) {
-		if ('\\' == *source) {
+		if (*source == '\\') {
 			quote = TRUE;	/* next char as-is */
-			if ('\\' == *++source) {
+			if (*++source == '\\') {
 				*ptr++ = *source++;
 				quote = FALSE;
 			}
 			continue;
 		}
-		if (('"' == *source) && !quote)
+		if ((*source == '"') && !quote)
 			break;	/* end of quoted-string */
 		*ptr++ = *source++;
 		quote = FALSE;
@@ -356,7 +388,7 @@ decode_value(
 
 	while (*rptr) {
 		if (*rptr == '%' && IS_XDIGIT(*(rptr + 1)) && IS_XDIGIT(*(rptr + 2))) {
-			*wptr++ = XVAL(*(rptr + 1)) << 4 | XVAL(*(rptr + 2));
+			*wptr++ = (char) (XVAL(*(rptr + 1)) << 4 | XVAL(*(rptr + 2)));
 			rptr += 3;
 		} else
 			*wptr++ = *rptr++;
@@ -399,7 +431,7 @@ skip_equal_sign(
 	if (!(source = skip_space(source)))
 		return NULL;
 
-	if ('=' != *source++)
+	if (*source++ != '=')
 		/* no equal sign, invalid header, stop parsing here */
 		return NULL;
 
@@ -447,7 +479,7 @@ parse_params(
 		}
 
 		/* RFC 2231 Parameter Value Continuations */
-		if ((contp = strchr(name, '*')) && *(contp + 1) && *(contp + 1) >= '0' && *(contp + 1) <= '9') {
+		if ((contp = strchr(name, '*')) && *(contp + 1) >= '0' && *(contp + 1) <= '9') {
 			idx = atoi(contp + 1);
 			*contp = '\0';
 		}
@@ -458,7 +490,7 @@ parse_params(
 		}
 
 		/* catch parameter value; may be surrounded by double quotes */
-		if ('"' == *param)	/* parse quoted-string */
+		if (*param == '"')	/* parse quoted-string */
 			param = get_quoted_string(param, &value);
 		else {
 			/* parse token */
@@ -480,9 +512,9 @@ parse_params(
 		content->params = ptr;
 
 		/* advance pointer to next parameter */
-		while ((*param) && (';' != *param))
+		while ((*param) && (*param != ';'))
 			param++;
-		if (';' == *param)
+		if (*param == ';')
 			param++;
 	}
 }
@@ -609,7 +641,7 @@ get_param(
 			 */
 			if (p_list->enc_fallback) {
 				tmpval = p_list->value;
-				if (*tmpval == '=' && *++tmpval && *tmpval == '?') {
+				if (*tmpval == '=' && *(++tmpval) == '?') {
 					if ((tmpval = rfc1522_decode(p_list->value))) {
 						free(p_list->value);
 						p_list->value = my_strdup(tmpval);
@@ -656,7 +688,7 @@ parse_content_type(
 		content->subtype = my_strdup("octet-stream");
 		return;
 	} else
-		content->type = i;
+		content->type = (unsigned int) i;
 
 	subtype = strtok(NULL, PARAM_SEP);
 	/* save new subtype, or use pre-initialised value "plain" */
@@ -704,10 +736,13 @@ parse_content_type(
 
 static unsigned int
 parse_content_encoding(
-	const char *encoding)
+	char *encoding)
 {
 	unsigned int i;
 
+	/* Remove comments and white space */
+	remove_cwsp(encoding);
+
 	for (i = 0; content_encodings[i] != NULL; ++i) {
 		if (strcasecmp(encoding, content_encodings[i]) == 0)
 			return i;
@@ -715,7 +750,7 @@ parse_content_encoding(
 
 	/*
 	 * TODO: check rfc - may need to switch Content-Type to
-	 * application/octet-steam where this header exists but is unparseable.
+	 * application/octet-steam where this header exists but is unparsable.
 	 *
 	 * RFC 2045 6.2:
 	 * Labelling unencoded data containing 8bit characters as "7bit" is not
@@ -944,15 +979,24 @@ parse_rfc822_headers(
 	hdr->ext = new_part(NULL);		/* Initialise MIME data */
 
 	while ((line = tin_fgets(from, TRUE)) != NULL) {
-		if (read_news_via_nntp && to)
+		if (read_news_via_nntp && to) {
 			fprintf(to, "%s\n", line);		/* Put raw data */
-
+#ifdef DEBUG
+			if ((debug & DEBUG_NNTP) && verbose > 1)
+				debug_print_file("NNTP", "<<<%s%s", logtime(), line);
+#endif /* DEBUG */
+		}
 		/*
 		 * End of headers ?
 		 */
 		if (line[0] == '\0') {
 			if (to)
 				hdr->ext->offset = ftell(to);	/* Offset of main body */
+
+			/* avoid null subject */
+			if (!hdr->subj)
+				hdr->subj = my_strdup("");
+
 			return 0;
 		}
 
@@ -1139,6 +1183,7 @@ parse_multipart_article(
 {
 	char *line;
 	char *ptr;
+	const char *bd;
 	int bnd;
 	int state = M_SEARCHING;
 	t_bool is_rfc822 = FALSE;
@@ -1152,8 +1197,13 @@ parse_multipart_article(
 		 */
 		bnd = boundary_check(line, artinfo->hdr.ext);
 
-		if (read_news_via_nntp)
+		if (read_news_via_nntp) {
 			fprintf(artinfo->raw, "%s\n", line);
+#ifdef DEBUG
+			if ((debug & DEBUG_NNTP) && verbose > 1)
+				debug_print_file("NNTP", "<<<%s%s", logtime(), line);
+#endif /* DEBUG */
+		}
 
 		artinfo->hdr.ext->line_count += count_lines(line);
 		if (show_progress_meter)
@@ -1172,8 +1222,10 @@ parse_multipart_article(
 			 * When we have reached the end boundary of the outermost envelope
 			 * just log any trailing data for the raw article format.
 			 */
-			if (boundary_cmp(line, get_param(artinfo->hdr.ext->params, "boundary")) == BOUND_END)
-				depth = 0;
+			if ((bd = get_param(artinfo->hdr.ext->params, "boundary")) != NULL) {
+				if (boundary_cmp(line, bd) == BOUND_END)
+					depth = 0;
+			}
 #if 0 /* doesn't count tailing lines after envelop mime part - correct but confusing */
 			if (read_news_via_nntp && depth == 0)
 				while ((line = tin_fgets(infile, FALSE)) != NULL)
@@ -1185,6 +1237,7 @@ parse_multipart_article(
 						fprintf(artinfo->raw, "%s\n", line);
 					artinfo->hdr.ext->line_count++;
 				}
+				return tin_errno | TIN_EOF;		/* Flag EOF */
 			}
 #endif /* 0 */
 			return tin_errno;
@@ -1206,8 +1259,11 @@ parse_multipart_article(
 
 			case M_HDR:
 				switch (bnd) {
-					case BOUND_START:	/* TODO: skip error message if not -DDEBUG? */
-						error_message(2, _(txt_error_mime_start));
+					case BOUND_START:
+#ifdef DEBUG
+						if (debug & DEBUG_MISC)
+							error_message(2, _(txt_error_mime_start));
+#endif /* DEBUG */
 						continue;
 
 					case BOUND_NONE:
@@ -1269,7 +1325,7 @@ parse_multipart_article(
 						curr_part->line_count++;
 						break;
 
-					case BOUND_START:		/* Start new attchment */
+					case BOUND_START:		/* Start new attachment */
 						if (is_rfc822) {
 							--depth;
 							rfc822_part->line_count--;
@@ -1305,9 +1361,16 @@ parse_normal_article(
 	char *line;
 
 	while ((line = tin_fgets(in, FALSE)) != NULL) {
-		if (read_news_via_nntp)
+		if (read_news_via_nntp) {
 			fprintf(artinfo->raw, "%s\n", line);
+#ifdef DEBUG
+			if ((debug & DEBUG_NNTP) && verbose > 1)
+				debug_print_file("NNTP", "<<<%s%s", logtime(), line);
+#endif /* DEBUG */
+		}
+
 		++artinfo->hdr.ext->line_count;
+
 		if (show_progress_meter)
 			progress(artinfo->hdr.ext->line_count);
 	}
@@ -1346,7 +1409,7 @@ dump_art(
 	t_param *pptr;
 	struct t_header note_h = art->hdr;
 
-	fprintf(stderr, "\nMain body\nMIME-Version: %d\n", note_h.mime);
+	fprintf(stderr, "\nMain body\nMIME-Version: %u\n", note_h.mime);
 	fprintf(stderr, "Content-Type: %s/%s\nContent-Transfer-Encoding: %s\n",
 		content_types[note_h.ext->type], note_h.ext->subtype,
 		content_encodings[note_h.ext->encoding]);
@@ -1392,10 +1455,10 @@ parse_rfc2045_article(
 	t_openartinfo *artinfo,
 	t_bool show_progress_meter)
 {
-	int ret;
+	int ret = ART_ABORT;
 
-	if (!infile || (read_news_via_nntp && !(artinfo->raw = tmpfile())))
-		return ART_ABORT;
+	if (read_news_via_nntp && !(artinfo->raw = tmpfile()))
+		goto error;
 
 	if (!read_news_via_nntp)
 		artinfo->raw = infile;
@@ -1405,6 +1468,12 @@ parse_rfc2045_article(
 	if ((ret = parse_rfc822_headers(&artinfo->hdr, infile, artinfo->raw)) != 0)
 		goto error;
 
+	/* no article data returned, just a '.' after 220er response */
+	if (artinfo->hdr.ext->offset == 0) {
+		ret = ART_UNAVAILABLE;
+		goto error;
+	}
+
 	/*
 	 * Is this a MIME article ?
 	 * We don't bother to parse all plain text articles
@@ -1414,8 +1483,10 @@ parse_rfc2045_article(
 			/* Strip off EOF condition if present */
 			if (ret & TIN_EOF) {
 				ret ^= TIN_EOF;
-				/* TODO: skip error message if not -DDEBUG? */
-				error_message(2, _(txt_error_mime_end), content_types[artinfo->hdr.ext->type], artinfo->hdr.ext->subtype);
+#ifdef DEBUG
+				if (debug & DEBUG_MISC)
+					error_message(2, _(txt_error_mime_end), content_types[artinfo->hdr.ext->type], artinfo->hdr.ext->subtype);
+#endif /* DEBUG */
 				if (ret != 0)
 					goto error;
 			} else
@@ -1463,10 +1534,11 @@ open_art_fp(
 		char buf[PATH_LEN];
 		char pbuf[PATH_LEN];
 		char fbuf[NAME_LEN + 1];
-		char group_path[PATH_LEN];
+		char *group_path = my_malloc(strlen(group->name) + 2); /* tailing "/\0" */;
 
 		make_group_path(group->name, group_path);
 		joinpath(buf, sizeof(buf), group->spooldir, group_path);
+		free(group_path);
 		snprintf(fbuf, sizeof(fbuf), "%"T_ARTNUM_PFMT, art);
 		joinpath(pbuf, sizeof(pbuf), buf, fbuf);
 
@@ -1483,7 +1555,7 @@ open_art_fp(
 /* ------------------------for accessing articles ------------------- */
 
 /*
- * Open's and postprocesses and article
+ * Opens and postprocesses an article
  * Populates the passed in artinfo structure if successful
  *
  * Returns:
@@ -1514,13 +1586,13 @@ art_open(
 	progress_mesg = pmesg;
 	if (parse_rfc2045_article(fp, art->line_count, artinfo, show_progress_meter) != 0) {
 		progress_mesg = NULL;
-		return ART_ABORT;
+		return ((tin_errno == 0) ? ART_UNAVAILABLE : ART_ABORT);
 	}
 	progress_mesg = NULL;
 
 	/*
 	 * TODO: compare art->msgid and artinfo->hdr.messageid and issue a
-	 *       warinng (once) about broken overviews if they differ
+	 *       warning (once) about broken overviews if they differ
 	 */
 
 	if ((artinfo->tex2iso = ((group->attribute->tex2iso_conv) ? is_art_tex_encoded(artinfo->raw) : FALSE)))
@@ -1535,7 +1607,7 @@ art_open(
 #endif /* DEBUG_ART */
 
 	/*
-	 * If Newsgroups is empty its a good bet the article is a mail article
+	 * If Newsgroups is empty it is a good bet the article is a mail article
 	 * TODO: Why do this ?
 	 */
 	if (!artinfo->hdr.newsgroups)
